package com.idtech.jpos;

import java.lang.reflect.*;
import jpos.*;
import jpos.loader.*;
import jpos.config.*;
import jpos.util.tracing.Tracer;
import jpos.util.*;

public class IDTECHJposServiceInstanceFactory extends Object
	implements JposServiceInstanceFactory
{
	public IDTECHJposServiceInstanceFactory() {}

	public JposServiceInstance createInstance(String logicalName, JposEntry entry)
        throws JposException
    {		
		if (!entry.hasPropertyWithName("serviceClass")) {
			throw new JposException(JposConst.JPOS_E_NOSERVICE,
                "The JposEntry does not contain the 'serviceClass' property");
		}
		JposServiceInstance serviceInstance = null;
	
		initTracer();

		try  {
			String serviceClassName =  (String) entry.getPropertyValue(
                                    "serviceClass");
        
			Class serviceClass = Class.forName(serviceClassName);
        
			Class[] cs = new Class[1];
			cs[0] =  JposEntry.class;
			Constructor ctor= serviceClass.getConstructor(cs);
        
			Object[] ps = new Object[1];
			ps[0] = entry;
			serviceInstance = (JposServiceInstance)ctor.newInstance(ps);
		} catch (Exception e) {
			System.err.print(e);
			throw new JposException(JposConst.JPOS_E_NOSERVICE,
                                "Could not create the service instance!", e);
		}
		return serviceInstance;
    }

	/** Initialize the current Tracer instance using the DefaultProperties class */
	private void initTracer()
	{
		JposProperties props = new DefaultProperties();
		props.loadJposProperties();

		if( !props.isPropertyDefined( JposProperties.TRACING_PROP_NAME ) )
			Tracer.getInstance().setOn( false );
		else
		{
			String tracingPropValue = props.getPropertyString( JposProperties.TRACING_PROP_NAME );

			if( tracingPropValue.equalsIgnoreCase( JposProperties.TRACING_ON_PROP_VALUE ) ||
					tracingPropValue.equalsIgnoreCase( JposProperties.TRACING_TRUE_PROP_VALUE ) ){
				Tracer.getInstance().setOn( true );
			}
		}
	}

}
