package com.idtech.service;

import com.idtech.connector.*;

import java.awt.Point;
import java.util.*;

import jpos.*;
import jpos.services.SignatureCaptureService113;
import jpos.events.*;
import jpos.services.EventCallbacks;
import jpos.util.tracing.Tracer;
import jpos.config.*;

public class uSignService implements SignatureCaptureService113,
									SignatureCaptureConst,
									JposConst,
									uSignEventListener
{
	protected uSignConnector connector;
	private   EventCallbacks eventCallbacks;
	private   Vector 	eventQueue;
	private   Vector    storePointQueue;
	private   JposEntry jposEntries;
	private   String    version;

	/* JPOS properties */
	private   boolean 	isOpened;
	private   boolean 	isClaimed;
	private   boolean   autoDisable;
	private   boolean 	deviceEnabled;
	private   boolean	dataEventEnabled;
	private   boolean   realTimeDataEnabled;
	private   boolean   m_bRealTimeDataEnabled;
	private	  boolean	freezeEvents;
	private	  int		state;
	private   boolean 	isBeginCapture;
	private   boolean   capDisplay;
	private   boolean   capRealTimeData;
	private   boolean   capUserTerminated;
	private   String 	checkHealthText;
	private   String	physicalDeviceDescription;
	private   java.awt.Point[] pointArray;
	private   byte[]    rawData;
	
	// Constructor Method called by instance factory to get settings.
	public uSignService(JposEntry entry) {
		Tracer.getInstance().println("Enter uSignService constructor\n");
		jposEntries = entry;
	}
	
	// Base service
	// Common Capabilities,Properties,Methods
	public void open(String logicalName, EventCallbacks cb) throws JposException 
	{
		if ( isOpened == true)
			throw new JposException (JPOS_E_FAILURE, "Service is already opened" );	
		// Initialize connector variables
		String cntorClass = (String)jposEntries.getPropertyValue("DeviceClass");
		String cntorName  = (String)jposEntries.getPropertyValue("DeviceName");
		String cntorSDK  = (String)jposEntries.getPropertyValue("SDK");
		if((cntorClass == null) || (cntorName == null) || (cntorSDK == null)) 
			throw new JposException(JPOS_E_FAILURE, "No connector class found in the jpos.xml");
		
		if(connector == null){
			connector = new uSignConnector();
		}
		
		if( ! connector.open( cntorClass, cntorName, cntorSDK ) )
			throw new JposException(JPOS_E_NOHARDWARE, "Fail to open connector," + cntorClass + ":" + cntorName);
		version = connector.getVersion();
		if( version == null ) {
			connector.close();
			connector = null;
			throw new JposException(JPOS_E_NOHARDWARE, "The device fail to response or no device.");		
		}
		
		eventCallbacks = cb;
		eventQueue = new Vector();
		storePointQueue = new Vector();
		// Initialize properties	
		isOpened = true;
		deviceEnabled = false;
		dataEventEnabled = false;
		realTimeDataEnabled = false;
		autoDisable = false;
		freezeEvents = false;
		isBeginCapture = false;
		capDisplay = true;
		capRealTimeData = true;
		capUserTerminated = false;
		idleState();
	}
	public void claim(int timeout) throws JposException 
	{
		if (isOpened == false){
			errorState();
			throw new JposException(JPOS_E_ILLEGAL, "Device is closed.");
		}
		if (isClaimed == true){
			errorState();
			throw new JposException(JPOS_E_ILLEGAL, "Device is already claimed.");
		}
		isClaimed = true;
		idleState();
		new Thread() {
    		public void run() {
    			while (isClaimed) {
    				try {
						Thread.sleep(10);
						fireEvent();
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
    			}
    		}
    	}.start();
	}
	public void release() throws JposException 
	{
		if ( isClaimed == false) 
			return ;
		if (isBeginCapture){
			this.endCapture();
		}
		if (deviceEnabled) {
			this.setDeviceEnabled(false);
		}
		eventQueue.removeAllElements();	
		storePointQueue.removeAllElements();
		isClaimed = false;		
		idleState();
	}
	public void close() throws JposException 
	{
		if(!isOpened) return;
		if (isClaimed) 
			release();
		connector.close();
		clearInputProperties();
		isOpened = false;
		closedState();
	}
	public boolean getClaimed() throws JposException 
	{
		return isClaimed;
	}
	public void setDeviceEnabled(boolean deviceEnabled) throws JposException 
	{
		if (isOpened == false)
			throw new JposException(JPOS_E_ILLEGAL, "Device is not opened.");
		if (getClaimed() == false)
			throw new JposException(JPOS_E_ILLEGAL, "Device is not claimed.");
		// deviceEnabled change false from true, stop the a going capture.
		if (this.deviceEnabled == true && deviceEnabled == false)
		{
			if(isBeginCapture == true)
				this.endCapture();	
		}
		if (capDisplay)
			connector.clearSignature();
		this.deviceEnabled = deviceEnabled;
		idleState();
	}
	public boolean getDeviceEnabled() throws JposException 
	{	
		return deviceEnabled;
	}
	public void setFreezeEvents(boolean fe) throws JposException 
	{
		freezeEvents = fe;
	}
	public boolean getFreezeEvents() throws JposException 
	{
		return freezeEvents;
	}
	public void checkHealth(int arg0) throws JposException {
		checkHealthText = "Complete checking health.";
		idleState();
	}
	public String getCheckHealthText() throws JposException {
		return checkHealthText;
	}
	public String getDeviceServiceDescription() throws JposException {
		return DEVICE_SERVICE_DESCRIPTION;
	}
	public int getDeviceServiceVersion() throws JposException {
		return DEVICE_SERVICE_VERSION;
	}
	public String getPhysicalDeviceDescription() throws JposException {
		physicalDeviceDescription = version == null ? this.getPhysicalDeviceName()+"SignatureCapture" : version;
		return physicalDeviceDescription;
	}
	public String getPhysicalDeviceName() throws JposException {
		return PHYSICAL_DEVICE_NAME;
	}
	public int getState() throws JposException {
		return state;
	}
	public void deleteInstance() throws JposException {
		idleState();
	}
	public void directIO(int arg0, int[] arg1, Object arg2)
			throws JposException {
		idleState();
	}
	
	// Capabilities12
	public boolean getCapDisplay() throws JposException
	{
		return capDisplay;
	}
    public boolean getCapRealTimeData() throws JposException
    {
		return capRealTimeData;
    }
    public boolean getCapUserTerminated() throws JposException
    {
		return capUserTerminated;
    }
    // Capabilities13
    public int     getCapPowerReporting() throws JposException
    {
		return JPOS_PR_NONE;
    }
    // Capabilities18
    public boolean getCapStatisticsReporting() throws JposException
    {
		return false;
    }
    public boolean getCapUpdateStatistics() throws JposException
    {
		return false;
    }
    // Capabilities19
    public boolean getCapCompareFirmwareVersion() throws JposException
    {
		return false;
    }
    public boolean getCapUpdateFirmware() throws JposException
    {
		return false;
    }
    
    // Properties12
    public boolean getAutoDisable() throws JposException
    {
		return autoDisable;
    }
    public void    setAutoDisable(boolean autoDisable) throws JposException
    {
    	this.autoDisable = autoDisable;
    }
    public int     getDataCount() throws JposException
    {
    	int sz, dc;
		Vector q;
		dc = 0;
		q = eventQueue;
		sz = q.size();
		for (int i = 0; i < sz; i ++)
		{
			if ( isDataEvent(q.elementAt(i)))
				dc ++;
		}
		return dc;
    }
    public boolean getDataEventEnabled() throws JposException
    {
		return dataEventEnabled;
    }
    public void    setDataEventEnabled(boolean dataEventEnabled)
                       throws JposException
    {
    	this.dataEventEnabled = dataEventEnabled;
    }
    public int     getMaximumX() throws JposException    
    {
		return MAXIMUM_X;
    }
    public int     getMaximumY() throws JposException
    {
		return MAXIMUM_Y;
    }
    public java.awt.Point[] getPointArray() throws JposException
    {
		return pointArray;
    }
    public byte[]  getRawData() throws JposException
    {
		return rawData;
    }
    public boolean getRealTimeDataEnabled() throws JposException
    {
		return realTimeDataEnabled;
    	
    }
    public void    setRealTimeDataEnabled(boolean realTimeDataEnabled)
                       throws JposException
    {
    	this.realTimeDataEnabled = realTimeDataEnabled;
    }
    // Properties13
    public int     getPowerNotify() throws JposException
    {
		return JPOS_PN_DISABLED;
    }
    public void    setPowerNotify(int powerNotify) throws JposException
    {
    	if (getCapPowerReporting() == JPOS_PR_NONE && powerNotify==JPOS_PN_ENABLED){
			errorState();
			throw new JposException(JPOS_E_ILLEGAL);
		}
    }
    public int     getPowerState() throws JposException
    {
		return JPOS_PS_UNKNOWN;
    }
    
    // Methods12
    public void    beginCapture(String formName) throws JposException
    {
    	if (!deviceEnabled)
			throw new JposException(JPOS_E_ILLEGAL, "Device is disabled.");
		if (isBeginCapture)
			throw new JposException(JPOS_E_BUSY, "Another Capture is beginning to runing.");
		if(connector.startCapture() == false)
			throw new JposException(JPOS_E_FAILURE, "Begin Capture failed");
		connector.removeuSignEventListener();
		connector.registeruSignEventListener(this);
		m_bRealTimeDataEnabled = realTimeDataEnabled;
		isBeginCapture = true;
    	busyState();
    }
    public void    endCapture() throws JposException
    {
    	if (!isBeginCapture) return;
	//	if(connector.exitCapture() == false)
	//		throw new JposException(JPOS_E_FAILURE, "End Capture failed");
    	connector.exitCapture();
    	connector.removeuSignEventListener();
    	if(!m_bRealTimeDataEnabled)
    	{
    		int len = storePointQueue.size();
    		if(len > 0)
    		{
	    		Vector  storePointArray = new Vector();
	    		Vector<Byte>  storerawData = new Vector<Byte>();
	    		for(int i = 0; i < len; i++)
	    		{
		    		JposEvent e = (JposEvent)storePointQueue.remove(0);
		    		if ( e instanceof uSignDataEvent ) {
		    			uSignDataEvent se = (uSignDataEvent) e;
		    			int m = se.getCapturePoinArray().length;
		    			int n = se.getCaptureRawData().length;
		    			for( int j = 0; j < m; j++ )
		    				storePointArray.add(new Point(se.getCapturePoinArray()[j]));
		    			for( int k = 0; k < n; k++ )
		    				storerawData.add((byte)se.getCaptureRawData()[k]);
		    		}
	    		}
	    		Point[] pointArrayTemp = new Point[storePointArray.size()];
	            for(int x = 0; x < storePointArray.size(); x++)
	            {
	                pointArrayTemp[x] = (Point)storePointArray.elementAt(x);
	            }
	            byte[] rawDataTemp = new byte[storerawData.size()];
	            for(int y = 0; y < storerawData.size(); y++)
	            {
	            	rawDataTemp[y] = (byte)storerawData.elementAt(y);
	            }
	            pointArray = pointArrayTemp;
	    		rawData = rawDataTemp;
	    		storePointArray.removeAllElements();
	    		storerawData.removeAllElements();
    		}
    		else
    		{
    			pointArray = null;
        		rawData = null;
    		}
    	}
    	else
    	{
    		storePointQueue.removeAllElements();
    		pointArray = null;
    		rawData = null;
    	}
    	isBeginCapture = false;
    	idleState();
    }
    public void    clearInput() throws JposException
    {
    	eventQueue.clear();
    }
    // Methods18
    public void    resetStatistics(String statisticsBuffer)
                       throws JposException
    {
    	errorState();
		throw new JposException(JPOS_E_FAILURE, "Unsupported operation: resetStatistics");
    }
    public void    retrieveStatistics(String[] statisticsBuffer)
                       throws JposException
    {
    	errorState();
		throw new JposException(JPOS_E_FAILURE, "Unsupported operation: retrieveStatistics");
    }
    public void    updateStatistics(String statisticsBuffer)
                       throws JposException
    {
    	errorState();
		throw new JposException(JPOS_E_FAILURE, "Unsupported operation: updateStatistics");
    }
    // Methods19
    public void    compareFirmwareVersion(String firmwareFileName, int[] result)
                       throws JposException
    {
    	errorState();
		throw new JposException(JPOS_E_FAILURE, "Unsupported operation: compareFirmwareVersion");
    }
    public void    updateFirmware(String firmwareFileName)
                       throws JposException
    {
    	errorState();
		throw new JposException(JPOS_E_FAILURE, "Unsupported operation: updateFirmware");
    }
    // Methods110
    public void    clearInputProperties()
                       throws JposException
    {
    	pointArray = null;
    	rawData = null;
    }
	
 // Implement uSignEventListener interface Methods
    public void captureDataEntry(int[] data, int len)
    { 
    	int length = len/5;
    	Point[] pointArrayTemp = new Point[length];
    	byte[] rawDataTemp = new byte[len];
    	pointArrayTemp = FBPtoPoints(data, len);
    	rawDataTemp = Convert(data, len);
    	if(m_bRealTimeDataEnabled)
    		eventQueue.add(new uSignDataEvent(this, 1 , pointArrayTemp, rawDataTemp));
    	else
    		storePointQueue.add(new uSignDataEvent(this, 1, pointArrayTemp, rawDataTemp));
    }
	public void captureError()
	{
		if(m_bRealTimeDataEnabled)
			eventQueue.add(new ErrorEvent(this, JPOS_E_EXTENDED, 0, JPOS_EL_INPUT, JPOS_ER_CLEAR));
		else
			storePointQueue.add(new ErrorEvent(this, JPOS_E_EXTENDED, 0, JPOS_EL_INPUT, JPOS_ER_CLEAR));
	}
    
    // Add Additional Methods and Class
    /** 
	 * state setting functions.
	 */
	private void idleState() {
		state = JPOS_S_IDLE;
	}
	private void errorState() {
		state = JPOS_S_ERROR;
	}
	private void busyState() {
		state = JPOS_S_BUSY;
	}
	private void closedState() {
		state = JPOS_S_CLOSED;
	}
	/**
	 * FBPtoPoints.
	 */
	private Point[] FBPtoPoints(int[] raw, int len)
	{
		int i=0, j=0, x=0, y=0;
		int len1 = (len/5);
		int[] buf = new int[4]; 
    	Point[] parr = new Point[len1];
		while (len >= 5) 
		{
			if(raw[i]==0x90)
			{
				buf[0] = (raw[i+2]&0x7f);
				buf[0] |= ((raw[i+1]&0x08)<<4);
				buf[1] = ((raw[i+1]&0x30)>>4);
				buf[2] = (raw[i+3]&0x7f);
				buf[2] |= ((raw[i+1]&0x01)<<7);
				buf[3] = ((raw[i+1]&0x06)>>1);
			}
			else if(raw[i]==0x80)
			{
				buf[0] = buf[1] = buf[2] = buf[3] = 0xff;
			}
			else
			{
				return null;
			}
			x = (buf[1]<<8)|buf[0];
			y = (buf[3]<<8)|buf[2];
			parr[j] = new Point(x, y);
			i += 5;
			j++;
			len -= 5;
		}
		return parr;
	}
	/**
	 * int convert to byte.
	 */
	private byte[] Convert(int[] src, int len)
	{	
		byte[] r = new byte[len];
		if (src == null || len == 0)
			return r;
		for ( int i = 0; i < len; i++) {
			r[i] = (byte) (src[i]);
		}
		return r;
	}
	/**
	 * check event type.
	 */
	private boolean isDataEvent(Object obj)	{		
		return obj instanceof DataEvent;
	}
	/**
	 * fire event in separate thread.
	 */
	private void fireEventThread() {		
		new Thread() {
			public void run() {
				fireEvent();
			}
		}.start();
	}
	/**
	 * check the condition and if possible fire ONE event or no more event.
	 */
	private void fireEvent() {
		
		if (freezeEvents == true) return ;
		if (dataEventEnabled == false) return;
		if (eventQueue.size() == 0) return ;

		JposEvent e = (JposEvent)eventQueue.remove(0);
		
		if ( e instanceof uSignDataEvent ) {
			uSignDataEvent se = (uSignDataEvent) e;
			pointArray = se.getCapturePoinArray();
			rawData = se.getCaptureRawData();
			eventCallbacks.fireDataEvent(se);
			return ;			
		}
		
		if ( e instanceof ErrorEvent) {
			eventCallbacks.fireErrorEvent((ErrorEvent)e);
			Tracer.getInstance().println("A ErrorEvent is fired");
			return ;
		}
		
	}
    
	private static String DEVICE_SERVICE_DESCRIPTION = "JPOS service for uSign300";
	private static int	DEVICE_SERVICE_VERSION = 1013000;
	private static String PHYSICAL_DEVICE_NAME = "uSign300";
	
	private static int MAXIMUM_X = 319;
	private static int MAXIMUM_Y = 239;
}
